// adapted from http://delphi.icm.edu.pl/ftp/d20free/dirmon.zip

// v2
// - added sender to event

unit ShellChange;

interface

uses classes, windows;

type
  TFileCreateEvent=procedure(sender:TObject; filename:string) of object;

  TShellChangeMonitor=class
  private
    FDirectoryHandle: THandle;
    FNotificationBuffer: array[0..4096] of Byte;
    FWatchThread: TThread;
    FNotifyFilter: DWORD;
    FOverlapped: TOverlapped;
    FPOverlapped: POverlapped;
    FBytesWritten: DWORD;
    FCompletionPort: THandle;
    FOnFileCreated: TFileCreateEvent;
    FDirectory: string;
    FSubTree: boolean;
    procedure DoOnFileCreated(filename:string);
  public
    procedure start;
    procedure stop;
    constructor Create;
    destructor Destroy; override;
  published
    property OnFileCreated:TFileCreateEvent read FOnFileCreated write FOnFileCreated;
    property Directory:string read FDirectory write FDirectory;
    property SubTree:boolean read FSubTree write FSubTree default true;
  end;

implementation

uses shlobj, activex, sysutils;

type
  PFileNotifyInformation = ^TFileNotifyInformation;
  TFileNotifyInformation = record
    NextEntryOffset: DWORD;
    Action: DWORD;
    FileNameLength: DWORD;
    FileName: array[0..0] of WideChar;
  end;

const
  FILE_LIST_DIRECTORY   = $0001;

type
  TWaitThread = class(TThread)
  private
    shellChangeMonitor: TShellchangeMonitor;
    procedure HandleEvent;
  protected
    procedure Execute; override;
  public
    constructor Create(scm: TShellchangeMonitor);
  end;

constructor TWaitThread.Create(scm: TShellchangeMonitor);
begin
  inherited Create(True);
  shellChangeMonitor := scm;
  FreeOnTerminate := False;
end;

procedure TWaitThread.HandleEvent;
var
  FileOpNotification: PFileNotifyInformation;
  Offset: Longint;
begin
  with shellChangeMonitor do
  begin
    Pointer(FileOpNotification) := @FNotificationBuffer[0];
    repeat
      Offset := FileOpNotification^.NextEntryOffset;
      try
        shellChangeMonitor.DoOnFileCreated(WideCharToString(@(FileOpNotification^.FileName)));
      except
        on e:exception do
          raise; // log it
      end;
      PChar(FileOpNotification) := PChar(FileOpNotification)+Offset;
    until Offset=0;
  end;
end;

procedure TWaitThread.Execute;
var
  numBytes: DWORD;
//  cbOffset: DWORD;
  CompletionKey: DWORD;
begin
  while not Terminated do
  try
    GetQueuedCompletionStatus( shellChangeMonitor.FCompletionPort, numBytes, CompletionKey, shellChangeMonitor.FPOverlapped, INFINITE);
    if CompletionKey <> 0 then
    begin
      Synchronize(HandleEvent);
      with shellChangeMonitor do
      begin
        FBytesWritten := 0;
        ZeroMemory(@FNotificationBuffer, SizeOf(FNotificationBuffer));
        ReadDirectoryChanges(FDirectoryHandle, @FNotificationBuffer, SizeOf(FNotificationBuffer), SubTree, FNotifyFilter, @FBytesWritten, @FOverlapped, nil);
      end;
    end
    else
      Terminate;
  except
    on e:exception do
      raise;// log it
  end;
end;

{ TShellchangeMonitor }

constructor TShellChangeMonitor.Create;
begin
  FCompletionPort := 0;
  FDirectoryHandle := 0;
  FPOverlapped := @FOverlapped;
  ZeroMemory(@FOverlapped, SizeOf(FOverlapped));
  SubTree:=true;
end;

destructor TShellChangeMonitor.Destroy;
begin
  stop;
  inherited;
end;

procedure TShellChangeMonitor.DoOnFileCreated(filename: string);
begin
  if assigned(FOnFileCreated) then
    FOnFileCreated(self, includetrailingpathdelimiter(Directory)+filename);
end;

procedure TShellchangeMonitor.start;
begin
  FNotifyFilter:=FILE_NOTIFY_CHANGE_FILE_NAME;
  FDirectoryHandle := CreateFile(PChar(Directory),
    FILE_LIST_DIRECTORY,
    FILE_SHARE_READ or FILE_SHARE_WRITE or FILE_SHARE_DELETE,
    nil,
    OPEN_EXISTING,
    FILE_FLAG_BACKUP_SEMANTICS or FILE_FLAG_OVERLAPPED,
    0);
  if FDirectoryHandle = INVALID_HANDLE_VALUE then
  begin
    FDirectoryHandle := 0;
    raise exception.create(SysErrorMessage(GetLastError));
  end;
  FCompletionPort := CreateIoCompletionPort(FDirectoryHandle, 0, Longint(pointer(self)), 0);
  ZeroMemory(@FNotificationBuffer, SizeOf(FNotificationBuffer));
  FBytesWritten := 0;
  if not ReadDirectoryChanges(FDirectoryHandle, @FNotificationBuffer, SizeOf(FNotificationBuffer), subTree, FNotifyFilter, @FBytesWritten, @FOverlapped, nil) then
  begin
    CloseHandle(FDirectoryHandle);
    FDirectoryHandle := 0;
    CloseHandle(FCompletionPort);
    FCompletionPort := 0;
    raise exception.create(SysErrorMessage(GetLastError));
  end;
  FWatchThread := TWaitThread.Create(self);
  TWaitThread(FWatchThread).Resume;
end;

procedure TShellChangeMonitor.stop;
begin
  if FCompletionPort = 0 then
    exit;
  PostQueuedCompletionStatus(FCompletionPort, 0, 0, nil);
  FWatchThread.Terminate;
  FWatchThread.WaitFor;
  freeandnil(FWatchThread);
  CloseHandle(FDirectoryHandle);
  FDirectoryHandle := 0;
  CloseHandle(FCompletionPort);
  FCompletionPort := 0;
end;

end.
